#!/usr/bin/python
import sys
import os
import shutil
import time
import ujson

from fcntl import fcntl, F_GETFL, F_SETFL

from .reply_helpers import (Reply, reply_fail, REPLY_INTERNAL_SERVER_ERROR_500,
                            REPLY_BAD_REQUEST_400, REPLY_NOT_FOUND_404,
                            REPLY_CONFLICT_409)
from ..com.fs import try_lock_pid_file, try_rm_pid_file
from .. import eventlog


class CgiError(RuntimeError):
    pass


class CgiNotFoundError(CgiError):
    pass


class CgiBadRequestError(CgiError):
    pass


class CgiConflictError(CgiError):
    pass


class CgiInternalServerError(CgiError):
    pass


class CgiPathInfo(list):
    def __init__(self, *args, **kwargs):
        list.__init__(self, *args, **kwargs)

    @staticmethod
    def unpack(raw_path_info):
        """
        :type raw_path_info: str
        :rtype: CgiPathInfo
        """
        return CgiPathInfo(raw_path_info.split('/')[1:])


class CgiRequest:
    def __init__(self, method, path_info, query_string, body):
        """
        :type method: str
        :type path_info: CgiPathInfo
        :type query_string: str | None
        :type body: str | None
        """
        self.method = method
        self.path_info = path_info
        self.query_string = query_string
        self.body = body

    @property
    def body_jnode(self):
        """
        :rtype: dict
        :raise CgiBadRequestError:
        """
        try:
            return ujson.loads(self.body)
        except ValueError as x:
            eventlog.log_e("fail: loads <{}>".format(x.message))
            raise CgiBadRequestError

    @staticmethod
    def _read_stdin():
        """
        :rtype: str | None
        :raises CgiError:
        """
        # make stdin a non-blocking file
        fd = sys.stdin.fileno()
        flags = fcntl(fd, F_GETFL)
        fcntl(fd, F_SETFL, flags | os.O_NONBLOCK)

        try:
            return sys.stdin.read()
        except Exception as x:
            eventlog.log_e("fail: read stdin <{}>".format(x.message))
            raise CgiError

        finally:
            fcntl(fd, F_SETFL, flags)

    @classmethod
    def unpack(cls, path_info_factory):
        """
        :type path_info_factory: (str) -> CgiPathInfo
        :rtype: Request

        :raises CgiError:
        """
        try:
            # read request
            env_method = os.environ['REQUEST_METHOD']
            env_path_info = os.environ.get('PATH_INFO') or None
            env_query_string = os.environ.get('QUERY_STRING') or None
            body = cls._read_stdin() or None

            eventlog.log_i("%s|%s|%s|%s" % (env_method, env_path_info, env_query_string, body))

            # unpack request
            path_info = path_info_factory(env_path_info or '')
            return CgiRequest(env_method, path_info, env_query_string, body)

        except CgiError:
            eventlog.log_e("fail: unpack query-args")
            raise

        except Exception as x:
            eventlog.log_e("fail: unpack query-args <{}>".format(x.message))
            raise CgiError


class CgiLock(object):
    def __init__(self, lock_file):
        """
        :type lock_file: str

        :raise CgiInternalServerError:
        :raise CgiConflictError:
        """
        self._lock_file = lock_file

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        if try_lock_pid_file(self._lock_file):
            answer = fn(*fn_args, **fn_kwargs)
            if not try_rm_pid_file(self._lock_file):
                eventlog.log_e("fail: rm pidfile <{}>".format(self._lock_file))
            return answer

        else:
            eventlog.log_e("fail: http409. lock <{}>".format(self._lock_file))
            raise CgiConflictError


class CgiBackup(object):
    def __init__(self, bak_dir, bak_files):
        """
        :type bak_dir: str
        :type bak_files: dict

        """
        self._bak_dir = bak_dir
        self._bak_files = bak_files

    def __call__(self, fn, *extra_args, **extra_kwargs):
        return lambda *fn_args, **fn_kwargs: self._exec(fn, *fn_args, **fn_kwargs)

    # noinspection PyBroadException
    def _exec(self, fn, *fn_args, **fn_kwargs):
        """
        :raise CgiInternalServerError:
        """
        bak_dir = '%s.%s' % (self._bak_dir, int(time.time() * 1000000))

        self._create_backup(bak_dir)
        try:
            answer = fn(*fn_args, **fn_kwargs)
            self._remove_backup(bak_dir)
            return answer

        except Exception:
            self._restore_backup(bak_dir)
            raise

    def _create_backup(self, bak_dir):
        """
        :type bak_dir: str
        :raise CgiInternalServerError:
        """
        try:
            os.makedirs(bak_dir)
            for (name, file_path) in self._bak_files.iteritems():
                shutil.copy(file_path, os.path.join(bak_dir, name))

        except Exception:
            eventlog.log_e("fail: http500")
            eventlog.log_x()
            raise CgiInternalServerError

    def _restore_backup(self, bak_dir):
        """
        :type bak_dir: str
        :raise CgiInternalServerError:
        """
        # noinspection PyBroadException
        try:
            for (name, file_path) in self._bak_files.iteritems():
                shutil.copy(os.path.join(bak_dir, name), file_path)

        except Exception:
            eventlog.log_x()

        self._remove_backup(bak_dir)

    @staticmethod
    def _remove_backup(bak_dir):
        """
        :type bak_dir: str
        """
        # noinspection PyBroadException
        try:
            shutil.rmtree(bak_dir)
        except Exception:
            eventlog.log_e("fail: http500")
            eventlog.log_x()


class CgiModule:
    def __init__(self, handler):
        """
        :type handler: () -> Reply
        """
        self._handler = handler

    def run(self):
        """
        :rtype: Reply
        """
        return self._handler()


def run_cgi(path_info_factory, module_factory):
    """
    :type path_info_factory: (str) -> CgiPathInfo
    :type module_factory: (CgiRequest) -> CgiModule
    """

    # noinspection PyBroadException
    try:
        request = CgiRequest.unpack(path_info_factory)
        module = module_factory(request)

        response = module.run()
        response.send()

    except CgiBadRequestError:
        Reply(REPLY_BAD_REQUEST_400).send()

    except CgiNotFoundError:
        Reply(REPLY_NOT_FOUND_404).send()

    except CgiConflictError:
        Reply(REPLY_CONFLICT_409).send()

    except CgiInternalServerError:
        Reply(REPLY_INTERNAL_SERVER_ERROR_500).send()

    except Exception:
        eventlog.log_x()
        Reply(REPLY_INTERNAL_SERVER_ERROR_500).send()

    sys.exit(0)


def run_cgi2(path_info_factory, module_factory):
    """
    :type path_info_factory: (str) -> CgiPathInfo
    :type module_factory: (CgiRequest) -> CgiModule
    """

    # noinspection PyBroadException
    try:
        request = CgiRequest.unpack(path_info_factory)
        module = module_factory(request)
        reply = module.run()

        envelope_reply = Reply(int(reply.status), {
            'status': int(reply.status),
            'result': reply.body
        })

        envelope_reply.send()

    except CgiBadRequestError:
        Reply(REPLY_BAD_REQUEST_400, reply_fail(REPLY_BAD_REQUEST_400)).send()

    except CgiNotFoundError:
        Reply(REPLY_NOT_FOUND_404, reply_fail(REPLY_NOT_FOUND_404)).send()

    except CgiConflictError:
        Reply(REPLY_CONFLICT_409, reply_fail(REPLY_CONFLICT_409)).send()

    except CgiInternalServerError:
        Reply(REPLY_INTERNAL_SERVER_ERROR_500, reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)).send()

    except Exception:
        eventlog.log_x()
        Reply(REPLY_INTERNAL_SERVER_ERROR_500, reply_fail(REPLY_INTERNAL_SERVER_ERROR_500)).send()

    sys.exit(0)
