import os
import errno
import shutil
import fcntl
import ujson
import semver

from .. import eventlog

_export_pil_image_fn = None


def _export_pil_image(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    def fn(rgba, jpeg_path):
        """
        :type rgba: PIL.Image.Image
        :type jpeg_path: str
        """
        rgba.save(jpeg_path, 'JPEG', quality=60)

    def fn4(rgba, jpeg_path):
        """
        :type rgba: PIL.Image.Image
        :type jpeg_path: str
        """
        rgb_pil_image = rgba.convert('RGB')
        rgb_pil_image.save(jpeg_path, 'JPEG', quality=60)

    global _export_pil_image_fn
    if not _export_pil_image_fn:
        import PIL
        if semver.compare(PIL.PILLOW_VERSION, "4.2.0") == -1:
            _export_pil_image_fn = fn

        else:
            _export_pil_image_fn = fn4

    return _export_pil_image_fn(rgba_pil_image, filepath)


class FsError(RuntimeError):
    pass


# noinspection PyBroadException
def try_lock_pid_file(pid_path):
    """
    :type pid_path: str
    :rtype: bool
    """
    try:
        for retry in range(2):
            try:
                # Request exclusive (EX) non-blocking (NB) advisory lock.
                fd = os.open(pid_path, os.O_CREAT | os.O_WRONLY)
                fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
                os.write(fd, str(os.getpid()))
                return True

            except os.error as x:
                if x.errno != errno.ENOENT:
                    raise

            try:
                os.makedirs(os.path.dirname(pid_path))
            except Exception:
                pass

    except Exception:
        pass

    return False


def try_rm_pid_file(pid_path):
    """
    :type pid_path: str
    :rtype: bool
    """
    try:
        os.remove(pid_path)
        return True
    except os.error:
        pass

    return False


def silent_mkdir_r(dir_path):
    """
    :type dir_path: str
    """
    if not dir_path:
        return

    try:
        os.makedirs(dir_path)
    except os.error as x:
        if x.errno != errno.EEXIST:
            eventlog.log_w('fail: mkdir {} <{}>'.format(dir_path, errno.errorcode[x.errno]))


def silent_rm_r(rm_path):
    if rm_path:
        try:
            os.remove(rm_path)
            return
        except EnvironmentError:
            pass

        try:
            shutil.rmtree(rm_path, ignore_errors=True)
        except EnvironmentError:
            eventlog.log_x()


def silent_mv(src, dst):
    """
    :type src: str
    :type dst: str

    :raise EnvironmentError:
    """
    try:
        shutil.move(src, dst)
    except EnvironmentError as x:
        if x.errno == errno.EEXIST:
            silent_rm_r(dst)
            silent_mv(src, dst)

        elif x.errno != errno.ENOENT:
            raise


def silent_link(src, dst):
    """
    :type src: str
    :type dst: str

    :raise EnvironmentError:
    """
    for _ in [1, 2]:
        try:
            os.link(src, dst)
            break
        except OSError as x:
            if x.errno == errno.EEXIST:
                silent_rm_r(dst)
                pass

            elif x.errno == errno.ENOENT:
                silent_mkdir_r(os.path.dirname(dst))
                pass


def edit_dir(target_dir, fn, *fn_args, **fn_kwargs):
    try_count = 2
    dir_present = True
    while try_count > 0 and dir_present:
        # noinspection PyBroadException
        try:
            fn(*fn_args, **fn_kwargs)
            break

        except IOError as x:
            if x.errno == errno.ENOENT:
                dir_present = False

            else:
                eventlog.log_w('fail: fn (%s)' % x)

        except Exception:
            eventlog.log_x()

        if not dir_present:
            silent_mkdir_r(target_dir)
            dir_present = True

        try_count -= 1

    if not try_count:
        raise RuntimeError


def write_json_file(jnode, filepath, pretty=False):
    """
    :type jnode: dict
    :type filepath: str
    :type pretty: bool
    """
    indent = 2 if pretty else 0
    for _ in [1, 2]:
        try:
            with open(filepath, 'w') as json_file:
                ujson.dump(jnode, json_file, ensure_ascii=False, indent=indent)
                break

        except IOError as x:
            if x.errno != errno.ENOENT:
                eventlog.log_e('fail: save {} <{}>'.format(filepath, errno.errorcode[x.errno]))
                break

        silent_mkdir_r(os.path.dirname(filepath))


def write_jpeg_file(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    for _ in [1, 2]:
        try:
            _export_pil_image(rgba_pil_image, filepath)
            break

        except IOError as x:
            if x.errno != errno.ENOENT:
                eventlog.log_e('fail: save {} <{}>'.format(filepath, errno.errorcode[x.errno]))
                break

        silent_mkdir_r(os.path.dirname(filepath))


def write_rgba_file(rgba_pil_image, filepath):
    """
    :type rgba_pil_image: PIL.Image.Image
    :type filepath: str
    """
    for _ in [1, 2]:
        try:
            with open(filepath, 'wb') as f:
                f.write(rgba_pil_image.tobytes())
                break

        except IOError as x:
            if x.errno != errno.ENOENT:
                eventlog.log_e('fail: save {} <{}>'.format(filepath, errno.errorcode[x.errno]))
                break

        silent_mkdir_r(os.path.dirname(filepath))


def ln_schema(dst_schema_path, ui_jnode_path):
    """
    :type dst_schema_path: str
    :type ui_jnode_path: str
    """
    try:
        with open(ui_jnode_path, 'r') as json_file:
            ui_jnode = ujson.load(json_file)

        src_schema_path = ui_jnode['ro']['schema-abspath']
    except Exception as x:
        eventlog.log_e("fail: load {} ({})".format(ui_jnode_path, x.message))
        raise

    ln_schema2(dst_schema_path, src_schema_path)


def ln_schema2(dst_schema_path, src_schema_path):
    """
    :type dst_schema_path: str
    :type src_schema_path: str
    """
    try:
        os.remove(dst_schema_path)
    except OSError as x:
        if x.errno != errno.ENOENT:
            eventlog.log_e('fail: rm {} ({})'.format(dst_schema_path, x.message))
            raise

    eventlog.log_i('exec: ln -s %s %s' % (src_schema_path, dst_schema_path))
    try:
        silent_mkdir_r(os.path.dirname(dst_schema_path))
        os.symlink(src_schema_path, dst_schema_path)
        eventlog.log_i('done: ln')

    except OSError as x:
        eventlog.log_e('fail: ln ({})'.format(x.message))
        raise


def create_jnode_from_default(def_jnode_filepath, jnode_filepath):
    """
    :type def_jnode_filepath: str
    :type jnode_filepath: str
    """
    if not def_jnode_filepath or not jnode_filepath:
        eventlog.log_d('fail: empty jnode filepath')
        raise RuntimeError

    if not os.path.exists(jnode_filepath):
        # noinspection PyBroadException
        try:
            eventlog.log_d('try: cp ui-jnode from %s to %s' % (def_jnode_filepath, jnode_filepath))
            with open(def_jnode_filepath, 'r') as json_file:
                def_ui_jnode = ujson.load(json_file)

            silent_mkdir_r(os.path.dirname(jnode_filepath))
            with open(jnode_filepath, 'w') as json_file:
                ujson.dump(def_ui_jnode, json_file)
            eventlog.log_d('done')

        except Exception:
            eventlog.log_d('fail: create ui-jnode from %s to %s' % (def_jnode_filepath, jnode_filepath))
            raise


def load_json(filepath):
    """
    :type filepath: str
    :rtype: dict | list | None
    """
    try:
        with open(filepath) as json_file:
            jnode = ujson.load(json_file)
            return jnode

    except IOError as x:
        eventlog.log_e('fail: IOError {} <{}, errno={}>'.format(filepath, x.message, x.errno))

    except ValueError as x:
        eventlog.log_e('fail: ValueError {} <{}>'.format(filepath, x.message))


def load_json2(filepath):
    """
    :type filepath: str
    :rtype: dict | list

    :raise FsError:
    """
    try:
        with open(filepath) as json_file:
            jnode = ujson.load(json_file)
            return jnode

    except IOError as x:
        eventlog.log_e('fail: IOError {} <{}>'.format(filepath, errno.errorcode[x.errno]))
        raise FsError

    except ValueError as x:
        eventlog.log_e('fail: ValueError {} <{}>'.format(filepath, x.message))
        raise FsError


def join_path(fs_path, *fs_paths):
    return os.path.join(fs_path, *fs_paths)


def get_df(fs_path):
    """
    Query disk free statistics about the given path.

    :rtype: float
    :return: [0.0; 1.0]
    """
    st = os.statvfs(fs_path)
    free = st.f_bavail * st.f_frsize
    total = st.f_blocks * st.f_frsize
    # used = (st.f_blocks - st.f_bfree) * st.f_frsize

    return free / float(total)
